@@ -6,11 +6,8 @@ import casadi as cs
 import numpy as npy
 from csnlp.solutions import Solution, subsevalf, DMStruct
 from csnlp.debug import NlpDebug
-from csnlp.util import (
-    cached_property, cache_clearer,
-    struct_symSX, dict2struct,
-    np_random
-)
+from csnlp.util.funcs import cached_property, cache_clearer, np_random
+from csnlp.util.data import struct_symSX, dict2struct
 
 
 '''This dictates the order for operations related to dual variables.'''
@@ -247,9 +244,9 @@ class Nlp:
     @cached_property
     def lam_all(self) -> Union[cs.SX, cs.MX]:
         '''Gets all the dual variables of the NLP scheme in vector form,
-        irrespective of redundant `lbx` and `ubx` multipliers. If 
-        `remove_redundant_x_bounds`, then this property is equivalent to 
-        the `lam` property. 
+        irrespective of redundant `lbx` and `ubx` multipliers. If
+        `remove_redundant_x_bounds`, then this property is equivalent to
+        the `lam` property.
 
         Note: The order of the dual variables can be adjusted via
         `_DUAL_VARIABLES_ORDER`.'''
@@ -352,7 +349,8 @@ class Nlp:
 
     @cache_clearer(variables, dual_variables, h_lbx, h_ubx, lam, lam_all)
     def variable(
-        self, name: str,
+        self,
+        name: str,
         shape: Tuple[int, int] = (1, 1),
         lb: Union[npy.ndarray, cs.DM] = -npy.inf,
         ub: Union[npy.ndarray, cs.DM] = +npy.inf
@@ -420,8 +418,14 @@ class Nlp:
         lhs: Union[npy.ndarray, cs.DM, cs.SX, cs.MX],
         op: Literal['==', '>=', '<='],
         rhs: Union[npy.ndarray, cs.DM, cs.SX, cs.MX],
+        soft: bool = False,
         simplify: bool = True
-    ) -> Union[Tuple[cs.SX, cs.SX], Tuple[cs.MX, cs.MX]]:
+    ) -> Union[
+        Tuple[cs.SX, cs.SX],
+        Tuple[cs.MX, cs.MX],
+        Tuple[cs.SX, cs.SX, cs.SX],
+        Tuple[cs.MX, cs.MX, cs.MX],
+    ]:
         '''Adds a constraint to the NLP problem, e.g., `lhs <= rhs`.
 
         Parameters
@@ -434,6 +438,12 @@ class Nlp:
             Operator relating the two terms.
         rhs : casadi.SX, MX, DM or numerical
             Symbolic expression of the right-hand term of the constraint.
+        soft : bool, optional
+            If `True`, then a slack variable with appropriate size is added to
+            the NLP to make the inequality constraint soft, and returned.
+            This slack is automatically lower-bounded by 0, but remember to
+            penalize its magnitude in the objective. Slacks are not supported
+            for equality constraints. Defaults to `False`.
         simplify : bool, optional
             Optionally simplies the constraint expression, but can be disabled.
 
@@ -444,12 +454,17 @@ class Nlp:
             `h(x,u) <= 0`.
         lam : casadi.SX, MX
             The symbol corresponding to the constraint's multipliers.
+        slack : casadi.SX, MX, optional
+            The slack variable in case of `soft=True`; otherwise, only a
+            2-tuple is returned.
 
         Raises
         ------
         ValueError
             Raises if there is already another constraint with the same name;
             or if the operator is not recognized.
+        NotImplementedError
+            Raises if a soft equality constraint is requested.
         TypeError
             Raises if the constraint is not symbolic.
         '''
@@ -465,6 +480,9 @@ class Nlp:
         if op == '==':
             is_eq = True
             lb = npy.zeros(npy.prod(shape))
+            if soft:
+                raise NotImplementedError(
+                    'Soft equality constraints are not currently supported.')
         elif op == '<=':
             is_eq = False
             lb = npy.full(npy.prod(shape), -npy.inf)
@@ -475,6 +493,10 @@ class Nlp:
         else:
             raise ValueError(f'Unrecognized operator {op}.')
 
+        if soft:
+            slack = self.variable(f'slack_{name}', expr.shape, lb=0)[0]
+            expr -= slack
+
         self._cons[name] = expr
         group, con, lam = \
             ('_g', '_lbg', '_lam_g') if is_eq else ('_h', '_lbh', '_lam_h')
@@ -489,7 +511,7 @@ class Nlp:
 
         if self._solver is not None:
             self.init_solver(self._solver_opts)  # resets solver
-        return expr, lam_c
+        return (expr, lam_c, slack) if soft else (expr, lam_c)
 
     def minimize(self, objective: Union[cs.SX, cs.MX]) -> None:
         '''Sets the objective function to be minimized.
@@ -603,13 +625,15 @@ class Nlp:
             self._lam_lbx, self._lam_ubx)
         new = cs.vertcat(p, sol['x'], lam_g, lam_h, lam_lbx, lam_ubx)
         get_value = partial(subsevalf, old=old, new=new)
-        return Solution(
+        solution = Solution(
             f=float(sol['f']),
             vars=vars_,
             vals=vals,
             stats=self._solver.stats().copy(),
             _get_value=get_value
         )
+        self._failures += int(not solution.success)
+        return solution
 
     def to_function(
         self,
